@@ -1,7 +1,7 @@
 /* pam_item.c */
 
 /*
- * $Id: pam_item.c,v 1.16 2008/03/03 13:23:45 t8m Exp $
+ * $Id$
  */
 
 #include "pam_private.h"
@@ -11,13 +11,15 @@
 #include <string.h>
 #include <syslog.h>
 
-#define TRY_SET(X, Y)                  \
-{                                      \
-    char *_TMP_ = _pam_strdup(Y);      \
-    if (_TMP_ == NULL && (Y) != NULL)  \
-	 return PAM_BUF_ERR;           \
-    free(X);                           \
-    (X) = _TMP_;                       \
+#define TRY_SET(X, Y)                      \
+{                                          \
+    if ((X) != (Y)) {		           \
+	char *_TMP_ = _pam_strdup(Y);      \
+	if (_TMP_ == NULL && (Y) != NULL)  \
+	    return PAM_BUF_ERR;            \
+	free(X);                           \
+	(X) = _TMP_;                       \
+    }					   \
 }
 
 /* functions */
@@ -76,8 +78,10 @@ int pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 	 * modules.
 	 */
 	if (__PAM_FROM_MODULE(pamh)) {
-	    _pam_overwrite(pamh->authtok);
-	    TRY_SET(pamh->authtok, item);
+	    if (pamh->authtok != item) {
+		_pam_overwrite(pamh->authtok);
+		TRY_SET(pamh->authtok, item);
+	    }
 	} else {
 	    retval = PAM_BAD_ITEM;
 	}
@@ -90,8 +94,10 @@ int pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 	 * modules.
 	 */
 	if (__PAM_FROM_MODULE(pamh)) {
-	    _pam_overwrite(pamh->oldauthtok);
-	    TRY_SET(pamh->oldauthtok, item);
+	    if (pamh->oldauthtok != item) {
+		_pam_overwrite(pamh->oldauthtok);
+		TRY_SET(pamh->oldauthtok, item);
+	    }
 	} else {
 	    retval = PAM_BAD_ITEM;
 	}
@@ -130,6 +136,8 @@ int pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 	break;
 
     case PAM_XAUTHDATA:
+	if (&pamh->xauth == item)
+	    break;
 	if (pamh->xauth.namelen) {
 	    _pam_overwrite(pamh->xauth.name);
 	    free(pamh->xauth.name);
@@ -143,7 +151,7 @@ int pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 	if ((pamh->xauth.name=_pam_strdup(pamh->xauth.name)) == NULL) {
 	    memset(&pamh->xauth, '\0', sizeof(pamh->xauth));
 	    return PAM_BUF_ERR;
-	}	
+	}
 	if ((pamh->xauth.data=_pam_memdup(pamh->xauth.data,
 	    pamh->xauth.datalen)) == NULL) {
 	    _pam_overwrite(pamh->xauth.name);
@@ -153,6 +161,10 @@ int pam_set_item (pam_handle_t *pamh, int item_type, const void *item)
 	}
 	break;
 
+    case PAM_AUTHTOK_TYPE:
+	TRY_SET(pamh->authtok_type, item);
+	break;
+
     default:
 	retval = PAM_BAD_ITEM;
     }
@@ -243,6 +255,10 @@ int pam_get_item (const pam_handle_t *pamh, int item_type, const void **item)
 	*item = &pamh->xauth;
 	break;
 
+    case PAM_AUTHTOK_TYPE:
+	*item = pamh->authtok_type;
+	break;
+
     default:
 	retval = PAM_BAD_ITEM;
     }
@@ -258,7 +274,8 @@ int pam_get_user(pam_handle_t *pamh, const char **user, const char *prompt)
 {
     const char *use_prompt;
     int retval;
-    struct pam_message msg,*pmsg;
+    struct pam_message msg;
+    const struct pam_message *pmsg;
     struct pam_response *resp;
 
     D(("called."));
@@ -324,8 +341,7 @@ int pam_get_user(pam_handle_t *pamh, const char **user, const char *prompt)
     resp = NULL;
 
     retval = pamh->pam_conversation->
-	conv(1, (const struct pam_message **) &pmsg, &resp,
-	     pamh->pam_conversation->appdata_ptr);
+	conv(1, &pmsg, &resp, pamh->pam_conversation->appdata_ptr);
 
     if (retval == PAM_CONV_AGAIN) {
 	/* conversation function is waiting for an event - save state */
