@@ -1,5 +1,5 @@
 /*
- * $Id: pam_modutil_getspnam.c,v 1.2 2007/08/30 04:00:39 vorlon Exp $
+ * $Id$
  *
  * This function provides a thread safer version of getspnam() for use
  * with PAM modules that care about this sort of thing.
@@ -104,7 +104,7 @@ pam_modutil_getspnam(pam_handle_t *pamh, const char *user)
                 break;
         }
 	
-	length <<= 2;
+	length <<= PWD_LENGTH_SHIFT;
 
     } while (length < PWD_ABSURD_PWD_LENGTH);
 
