@@ -1,7 +1,7 @@
 /* pam_session.c - PAM Session Management */
 
 /*
- * $Id: pam_session.c,v 1.5 2006/01/12 10:06:49 t8m Exp $
+ * $Id$
  */
 
 #include "pam_private.h"
@@ -22,7 +22,7 @@ int pam_open_session(pam_handle_t *pamh, int flags)
     }
     retval = _pam_dispatch(pamh, flags, PAM_OPEN_SESSION);
 
-#if HAVE_LIBAUDIT
+#ifdef HAVE_LIBAUDIT
     retval = _pam_auditlog(pamh, PAM_OPEN_SESSION, retval, flags);
 #endif                                                                                
     return retval;
@@ -43,7 +43,7 @@ int pam_close_session(pam_handle_t *pamh, int flags)
 
     retval = _pam_dispatch(pamh, flags, PAM_CLOSE_SESSION);
 
-#if HAVE_LIBAUDIT
+#ifdef HAVE_LIBAUDIT
     retval = _pam_auditlog(pamh, PAM_CLOSE_SESSION, retval, flags);
 #endif
 
