@@ -403,9 +403,13 @@ last_login_failed(pam_handle_t *pamh, int announce, const char *user, time_t llt
     /* obtain the failed login attempt records from btmp */
     fd = open(_PATH_BTMP, O_RDONLY);
     if (fd < 0) {
+        int save_errno = errno;
 	pam_syslog(pamh, LOG_ERR, "unable to open %s: %m", _PATH_BTMP);
 	D(("unable to open %s file", _PATH_BTMP));
-	return PAM_SERVICE_ERR;
+        if (save_errno == ENOENT)
+	  return PAM_SUCCESS;
+	else
+	  return PAM_SERVICE_ERR;
     }
 
     while ((retval=pam_modutil_read(fd, (void *)&ut,
@@ -454,7 +458,7 @@ last_login_failed(pam_handle_t *pamh, int announce, const char *user, time_t llt
 		goto cleanup;
 	    }
 	}
-	
+
 	if (line != NULL || date != NULL || host != NULL) {
 	    /* TRANSLATORS: "Last failed login: <date> from <host> on <terminal>" */
 	    pam_info(pamh, _("Last failed login:%s%s%s"),
@@ -471,7 +475,7 @@ last_login_failed(pam_handle_t *pamh, int announce, const char *user, time_t llt
 		failed),
 	    failed);
 #else
-	if (daysleft == 1)
+	if (failed == 1)
 	    retval = asprintf(&line,
 		_("There was %d failed login attempt since the last successful login."),
 		failed);
